/*
 * bus.c - taiwins server bus implementation
 *
 * Copyright (c) 2019 Xichen Zhou
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/eventfd.h>
#include <tdbus.h>
#include <ctypes/helpers.h>

#include <taiwins/objects/utils.h>
#include "utils.h"

static struct tw_bus {
	struct wl_display *display;
	struct tdbus *dbus;
	struct wl_event_source *source;

	struct wl_listener display_distroy_listener;
} s_bus;

static inline struct tw_bus *
get_bus(void)
{
	return &s_bus;
}

static void
tw_bus_end(struct wl_listener *listener, void *data)
{
	struct tw_bus *bus = container_of(listener, struct tw_bus,
	                                  display_distroy_listener);
	struct tdbus *dbus = bus->dbus;

	if (bus->source)
		wl_event_source_remove(bus->source);

	tdbus_delete(dbus);
}

static int tw_bus_read_request(const struct tdbus_method_call *call)
{
	char *msg_received = NULL, msg_reply[128];
	const char *invalid_reply = "message not vaild";
	struct tdbus_message *reply;
	struct tdbus *bus = call->bus;

	tdbus_read(call->message, "%s", &msg_received);
	if (!msg_received || strlen(msg_received) > 120)
		goto err_call;

	strcpy(msg_reply, "recv: ");
	strcat(msg_reply, msg_received);
	free(msg_received);

	reply = tdbus_reply_method(call->message, NULL);
	tdbus_write(reply, "%s", msg_reply);
	tdbus_send_message(bus, reply);

	return 0;
err_call:
	if (msg_received)
		free(msg_received);
	reply = tdbus_reply_method(call->message, invalid_reply);
	tdbus_write(reply, "%s", invalid_reply);
	tdbus_send_message(bus, reply);
	return 0;
}

static struct tdbus_call_answer tw_bus_answer = {
	.interface = "org.taiwins.example",
	.method = "Hello",
	.in_signature = "s",
	.out_signature = "s",
	.reader = tw_bus_read_request,
};

struct tw_bus *
tw_bus_create_global(struct wl_display *display)
{
	struct tw_bus *bus = get_bus();

	bus->dbus = tdbus_new_server(SESSION_BUS, "org.taiwins");
	bus->display = display;

	if (!bus->dbus)
		return NULL;

	tw_set_display_destroy_listener(display,
	                                &bus->display_distroy_listener,
	                                tw_bus_end);
	bus->source = tw_bind_tdbus_for_wl_display(bus->dbus, display);

	tdbus_server_add_methods(bus->dbus, "/org/taiwins", 1, &tw_bus_answer);

	return &s_bus;
}
